<?php
/**
 * @file
 * Drupal Inspect module base class
 */

/**
 * Inspect base class
 * @uses PHP:ini:precision
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect */
class InspectBs {
  /**
   * Prevents erratic instantiation of this entirely static class
   */
  private function  __construct() {}

  /**
   * Maximum sub var recursion depth.
   *
   * @type integer
   */
  const DEPTH_MAX = 10;

  /**
   * Default sub var recursion depth.
   *
   * @type integer
   */
  const DEPTH_DEFAULT = 10;

  /**
   * Minimum string truncation.
   *
   * @type integer
   */
  const TRUNCATE_MIN = 10000;

  /**
   * Default string truncation.
   *
   * @type integer
   */
  const TRUNCATE_DEFAULT = 100;

  /**
   * Base 62 uniqid() if session started, otherwise empty.
   *
   * @var string $_sessionNo
   */
  protected static $_sessionNo = '';

  /**
   * Request number in current session.
   *
   * @var string $_requestNo
   */
  protected static $_requestNo = 0;

  /**
   * Log number in current request.
   *
   * @var integer $_logNo
   */
  protected static $_logNo = 0;

  /**
   * Default inspection options.
   *
   *   String options:
   *   - (array) needles: list of char|string needles to be used in str_replace
   *   - (array) replacers: list of char|string replacers to be used in str_replace
   *   - (string) quote_begin: char|string used as initial quote
   *   - (string) quote_end: char|string used as final quote
   *
   *   Formatting options:
   *   - (string) delimiter: between object/array buckets, and for InspectTrc between frames and frame buckets
   *   - (string) indent: indentation of object/array buckets
   *   - (string) enclose_tag_get: html enclosure tag when getting
   *   - (string) enclose_tag_log: html enclosure tag when logging
   *
   * @var array $_optionsDefault
   */
  protected static $_optionsDefault = array(
    'needles' => array( "\n", "\r", '<', '>', '"', "'" ),
    'replacers' => array( '_NL_', '_CR_', '&#60;', '&#62;', '&#34;', '&#39;' ),
    'quote_begin' => '&#62;&#62;',
    'quote_end' => '&#60;&#60;',
    'delimiter' => "\r",
    'indent' => '.  ',
    'enclose_tag_get' => 'pre',
    'enclose_tag_log' => 'pre',
  );

  /**
   * Whether server OS is *nix (non-windows, really).
   *
   * @var integer $_os_nix
   */
  protected static $_os_nix = -1; // never checked

  /**
   * Whether the mb_strlen function exists (PHP multibyte extension).
   *
   * @var integer $_mb_strlen_exists
   */
  protected static $_mb_strlen_exists = -1; // never checked

  /**
   * Document root of the current site, without trailing slash, and as regular expression (^/...).
   *
   * If OS isnt *nix the value is an array: first bucket forward slashed, second bucket backslashed.
   *
   * @var string|array $_docRoot
   */
  protected static $_docRoot = ''; // never checked

  /**
   * Length of document root.
   *
   * @var integer $_docRootLength
   */
  protected static $_docRootLength; // never checked

  /**
   * Cache of the module permissions.
   *
   * @var array $_permissions
   */
  private static $_permissions = array(
    'inspect log' => FALSE,
    'inspect log scalars' => FALSE,
    'inspect log paths' => FALSE,
    'inspect get' => FALSE,
    'inspect get scalars' => FALSE,
  );

  /**
   * _inspect() returns empty if this isnt truthy.
   *
   * @var bool $_initialized
   */
  private static $_initialized = FALSE;

  /**
   * Checks permissions and that all inspection options exist, and adds/checks the buckets depth and truncate according to permissions.
   *
   * Any method calling InspectBs::_inspect() must guarantee to call this method at least once before calling _inspect().
   *
   * If the calling method wants to call _inspect() more than once it must guarantee to re-use the return value of this method
   * for the later calls to _inspect().
   *
   * Also checks if the mb_strlen function exists.
   * @see InspectBs::$_optionsDefault
   * @param string $target
   *   - allowed values: get|log
   * @param array|integer|mixed $depth_or_options
   *   - see InspectBs::$_optionsDefault for supported options (plus depth and truncate)
   *   - non-array and non-NULL: is cast to integer (so '1' will work just fine)
   *   - if integer: it counts as suggested depth
   * @return array|bool|NULL
   *   - FALSE if the user isnt allowed to get|log inspections
   *   - NULL if arg $target isnt get or log
   */
  protected static function _checkPermissionsAndOptions($target, $depth_or_options = NULL) {
    $allow =& self::$_permissions;
    //  To avoid having an init method, we do init-stuff here.
    if (!self::$_initialized) {
      if (!array_key_exists('user', $GLOBALS) || !is_object($GLOBALS['user']) || !property_exists($GLOBALS['user'], 'uid')) {
        return FALSE;
      }
      if (user_access('inspect get')
          //  ...And isnt admin user 1 (or is user 1 but not prohibited to do get).
          && ($GLOBALS['user']->uid != 1 || !variable_get('inspect_disallow_get_admin', FALSE))) {
        $allow['inspect get'] = TRUE;
        if (user_access('inspect get scalars')) {
          $allow['inspect get scalars'] = TRUE;
        }
      }
      if (user_access('inspect log')) {
        $allow['inspect log'] = TRUE;
        if (user_access('inspect log scalars')) {
          $allow['inspect log scalars'] = TRUE;
        }
        if (user_access('inspect log paths')) {
          $allow['inspect log paths'] = TRUE;
        }
      }
      //  Check OS.
      $osNix = self::$_os_nix = strpos(PHP_OS, 'WIN') !== 0 ? 1 : 0;
      //  Check whether the mb_strlen function exists.
      self::$_mb_strlen_exists = function_exists('mb_strlen') ? 1 : 0;
      //  Find document root, for removal from absolute paths.
      self::$_docRootLength = strlen(
          $docRoot = substr(dirname(__FILE__), 0, -1 * (1 + strlen(drupal_get_path('module', 'inspect'))))
        );
      self::$_docRoot = $osNix ?
          ('/^' . preg_quote($docRoot, '/') . '/') :
          array(
            '/^' . preg_quote( str_replace('\\', '/', $docRoot), '/') . '/',  // all /
            '/^' . preg_quote( str_replace('/', '\\', $docRoot), '/') . '/'  // all \
          );
      //  Set/get session and request number.
      if (drupal_session_started()) {
        $found = false;
        $a =& $_SESSION;
        if (array_key_exists('module', $a)) {
          $a =& $a['module'];
          if (array_key_exists('inspect', $a)) {
            $a =& $a['inspect'];
            if (array_key_exists('sessionNo', $a)) {
              $found = true;
              self::$_sessionNo = $a['sessionNo'];
              self::$_requestNo = (++$a['requestNo']);
            }
          }
          else {
            $a['inspect'] = array();
          }
        }
        else {
          $a['module'] = array('inspect' => array());
        }
        if (!$found) {
          $a =& $_SESSION['module']['inspect'];
          $comps = explode('.', uniqid('', true));
          self::$_sessionNo = $a['sessionNo'] = self::baseConvert($comps[0], 16) . self::baseConvert($comps[1]);
          self::$_requestNo = $a['requestNo'] = 1;
        }
        unset($a);
      }
      self::$_initialized = TRUE;
    }
    //  Check general permissions.
    $hideScalars = FALSE;
    $hidePaths = FALSE;
    if ($target === 'get') {
      if (!$allow['inspect get']) {
        return FALSE;
      }
      $hidePaths = TRUE; // we always hide paths when getting
    }
    elseif ($target === 'log') {
      if (!$allow['inspect log']) {
        return FALSE;
      }
    }
    else {
      return NULL;
    }
    //  Check/load options.
    $opts = $depth_or_options;
    $depthOpt = static::DEPTH_DEFAULT;
    $truncOpt = static::TRUNCATE_DEFAULT;
    $catOpt = 'inspect';
    $filterOpt = NULL;
    $optsDef =& static::$_optionsDefault;
    if (!is_array($opts)) {
      if (!is_null($opts)) {
        $depthOpt = (int)$opts;
      }
      $opts = $optsDef; // copy
    }
    else {
      foreach ($optsDef as $k => $v) {
        if (!isset($opts[ $k ])) {
          $opts[ $k ] = $v;
        }
      }
      //  Make sure string needles and replacers arent falsy.
      if (!$opts['needles']) {
        $opts['needles'] = $optsDef['needles'];
      }
      if (!$opts['replacers']) {
        $opts['replacers'] = $optsDef['replacers'];
      }
      //  Read depth, hide scalars, truncate, hide paths, if they exist.
      if (isset($opts['depth'])) {
        $depthOpt = (int)$opts['depth'];
      }
      if (!$hideScalars && isset($opts['hide_scalars']) && $opts['hide_scalars']) {
        $hideScalars = TRUE;
      }
      if (isset($opts['truncate'])) {
        $truncOpt = (int)$opts['truncate'];
      }
      if (isset($opts['category']) && strlen($opts['category'])) {
        $catOpt = $opts['category'];
      }
      if (isset($opts['filter'])) {
        if (is_array($opts['filter'])) {
          if (count($opts['filter'])) {
            $filterOpt = $opts['filter'];
          }
        }
        elseif (strlen('' . $opts['filter'])) {
          $filterOpt = array(0 => $opts['filter']);
        }
      }
      //  We set the bucket hide_paths no matter what, and it can only override in negative (safer) manner.
      if (!$hidePaths && isset($opts['hide_paths']) && $opts['hide_paths']) {
        $hidePaths = TRUE;
      }
    }
    //  Check permission for describing scalar variable values.
    if ($target == 'get') {
      if (!$allow['inspect get scalars']) {
        $hideScalars = TRUE;
      }
    }
    else { // log
      if (!$allow['inspect log scalars']) {
        $hideScalars = TRUE;
      }
      if (!$allow['inspect log paths']) {
        $hidePaths = TRUE;
      }
    }
    //  Check overrider parameters and overall max/min.
    if ($depthOpt < 0) {
      $depthOpt = 0;
    }
    elseif ($depthOpt && $depthOpt > static::DEPTH_MAX) {
      $depthOpt = $depth_max;
    }
    //  If suggested truncate is more than overall min.
    if ($truncOpt < 0) {
      $truncOpt = 0;
    }
    elseif ($truncOpt > 0 && $truncOpt > static::TRUNCATE_MIN) {
      $truncOpt = static::TRUNCATE_MIN;
    }
    //  Load resolved values of depth and truncate into the options array.
    $opts['depth'] = $depthOpt;
    $opts['hide_scalars'] = $hideScalars;
    $opts['truncate'] = $truncOpt;
    $opts['category'] = $catOpt;
    $opts['filter'] = $filterOpt;
    $opts['hide_paths'] = $hidePaths;
    //  Flag whether getting or logging.
    //$opts['target'] = $target; - not used currently
    return $opts;
  }

  /**
   * @return string|integer
   */
  protected static function _logNumber() {
    if(self::$_sessionNo) {
      return self::$_sessionNo . ':' . self::$_requestNo . ':' . (++self::$_logNo);
    }
    return (++self::$_logNo);
  }

  /**
   * The actual variable analyser - do not call it with an options arg that hasnt been checked by _checkPermissionsAndOptions()
   *
   * See InspectVar::get() for specifics of arg $options.
   *
   *   String truncation:
   *   - truncate option less than one (hide strings) doesnt apply to strings that solely consist of digits (stringed integers)
   *   - a string bucket of an object/array keyed pass or password (lowercase) will always be hidden (no matter it's content)
   *
   *   Removes (what looks like) paths when getting (not logging), relative as well as absolute path:
   *   - if *nix server OS: only strings having no space(s), unix style absolute paths, and only looks for forward slash
   *   - if windows server OS: windows style absolute paths, and looks for forward and backward slashes
   *   - path removal takes place before escaping and truncating (except for overly long strings)
   *
   *   Reports non-integer numbers:
   *   - double as float
   *   - NaN as NaN and infinite as infinite (unless hiding scalars, in such case integer|float)
   *
   *   Reports string lengths and path removal (path reduced to filename):
   *   - un-truncated string, and when hiding scalar values: (multibyte length|ascii length)
   *   - truncated string: (multibyte length|ascii length|truncation length)
   *   - if path removed: there will be a fourth flag (exclamation mark), if no truncation truncation length will be hyphen
   *
   *   Recursive:
   *   - always stops at 10th (static::DEPTH_MAX) recursion or less (~ the depth option)
   *   - when buckets of an object/array is analyzed, buckets of buckets will be analyzed to one lesser depth
   *   - detects if a bucket of an object references the object itself (marked *RECURSION*)
   *   - detects GLOBALS array's self-referencing GLOBALS bucket (marked *RECURSION*)
   *   - doesnt check identity for arrays (not possible in PHP), nor 'same-ness' (too heavy, and prone to err)
   * @see InspectVar::get()
   * @param mixed $var
   * @param array &$options
   *   - by reference
   * @param integer $_curDepth
   *   - default: zero
   *   - recursion counter, do not call it with anything but zero
   * @return string
   */
  protected static function _inspect($var, &$options, $_curDepth = 0) {
    if (!self::$_initialized) {
      return '';
    }
    $maxDepth = $options['depth'];
    //  Recursion limit check.
    if ($_curDepth
        && $_curDepth > $maxDepth) {
      return '_INSPECT_DEPTH_LIMIT_:' . $maxDepth;
    }
    //  Object or array.
    $tArr = FALSE;
    if (is_object($var) || ($tArr = is_array($var))) {
      $s = '(' . ($tArr ? 'array' : ($tObj = get_class($var))) . ':';
      //  If at max depth: simply get length of the container.
      if ($_curDepth == $maxDepth) {
        if ($tArr) {
          $s .= count($var) . ')';
        }
        else { // stdClass cannot be counted, simply iterate
          $le = 0;
          foreach ($var as $sub) {
            ++$le;
          }
          $s .= $le . ')';
        }
        return $s;
      }
      //  Dive into container buckets.
      else {
        $delimIndent = ($delim = $options['delimiter'])
            . str_repeat($indent = $options['indent'], $_curDepth + 1);
        $le = 0;
        $sSub = '';
        //  Different iterations for object vs array, because:
        //  - array ust iterate by explicit reference,
        //    whereas object may fail fatally doing that (An iterator cannot be used with foreach by reference)
        //  - they must have different self reference checks
        //  - array must be checked for (all) numeric indices
        if ($tArr) {
          $numArr = TRUE; // if all numeric indices, use [] instead of {}
          foreach ($var as $key => &$sub) { // by explicit reference
            ++$le;
            //  GLOBALS check, skip self reference.
            if ($key === 'GLOBALS' && is_array($sub) && array_key_exists('GLOBALS', $sub)) {
              $sSub .= $delimIndent . 'GLOBALS: (array) *RECURSION*';
              $numArr = FALSE;
              continue;
            }
            if ($numArr && $key !== $le - 1) { // numeric array check
              $numArr = FALSE;
            }
            if ($options['filter'] && in_array($key, $options['filter'], TRUE)) {
              $sSub .= $delimIndent . $key . ': F';
            }
            else {
              $options['key'] = $key; // pass key, for password check
              $sSub .= $delimIndent . $key . ': ' . self::_inspect($sub, $options, $_curDepth + 1); // recursion
            }
          }
          return $s . $le . ') '
              . ($numArr ? '[' : '{')
              . $sSub . $delim . str_repeat($indent, $_curDepth)
              . ($numArr ? ']' : '}');
        }
        else { // object
          foreach ($var as $key => $sub) { // not &$sub, iterating object is implicitly by reference
            ++$le;
            if ($sub === $var) { // check for identity (self reference)
              $sSub .= $delimIndent . $key . ': (' . $tObj . ') *RECURSION*';
              continue;
            }
            if ($options['filter'] && in_array($key, $options['filter'], TRUE)) {
              $sSub .= $delimIndent . $key . ': F';
            }
            else {
              $options['key'] = $key; // pass key, for password check
              $sSub .= $delimIndent . $key . ': ' . self::_inspect($sub, $options, $_curDepth + 1); // recursion
            }
          }
          return $s . $le . ') {'
              . $sSub . $delim . str_repeat($indent, $_curDepth)
              . '}';
        }
      }
    }
    //  Scalars, NULL and resource.
    switch (($t = gettype($var))) {
      case 'boolean':
        return '(' . $t . ')' . (!$options['hide_scalars'] ? ($var ? ' TRUE' : ' FALSE') : '');
      case 'integer':
      case 'double':
      case 'float':
        if (!$options['hide_scalars'] && !is_finite($var)) {
          return '(' . (is_nan($var) ? 'NaN' : 'infinite') . ')';
        }
        return '(' . ($t == 'double' ? 'float' : $t) . ')'
            . ( $options['hide_scalars'] ? '' :
                (!$var ? ' 0' : (' ' . self::numberToString($var)) ) );
      case 'string':
        //  (string:n0|n1|n2): n0 ~ multibyte length, n1 ~ ascii length, n2 only occurs if when truncation
        //  Get for multibyte length and ascii length.
        $s = '(string:'
            . (
              !($le = strlen($var)) ? '0|0' :
                  ( (self::$_mb_strlen_exists > 0 ? mb_strlen($var) : '?') . '|' . $le )
            );
        //  If empty or hide scalar values.
        if ($options['hide_scalars'] || !$le) {
          return $s . ')'
              . ( $options['hide_scalars'] ? '' : // no quotes if hide value
                  (' ' . $options['quote_begin'] . $options['quote_end']) );
        }
        //  If very long string, truncate to absolute max before working on the string internals.
        $truncedTo = 0;
        if ($le > ($trunc = $options['truncate']) && $le > ($trunc_min = static::TRUNCATE_MIN)) {
          $var = substr($var, 0, $trunc_min);
          $truncedTo = $le = $trunc_min;
        }
        //  If truncate is zero.
        if (($pw = ($_curDepth > 0 && ($options['key'] === 'pass' || $options['key'] === 'password'))) || $trunc < 1) {
          //  Look for stringed integer (consists solely of digits), but only if not truncated.
          if (!$pw && !$truncedTo && preg_match('/^\d+$/', $var)) {
            return $s . ') ' . $options['quote_begin'] . $var . $options['quote_end'];
          }
          return $s . '|0)'; // |0 ~ flag truncation
        }
        //  Remove document root (always) and/or path.
        //  Do it before escaping (needles may contain dot or slash).
        $pathRem = FALSE;
        if ($le > 3) {
          if ($options['hide_paths']) { //  This algo also removes document root.
            if (self::$_os_nix) { // Non-windows.
              if (strpos($var, ' ') === FALSE // If no space in string.
                  && ($posSlash = strpos($var, '/') ) !== FALSE) { // And has forward slash.
                if (!$posSlash // Slash at pos zero ~ absolute path.
                    || ( $var{0} === '.'
                        && (strpos($var, '../') === 0 || strpos($var, './') === 0 ) ) // Relative.
                ) {
                  $pathRem = TRUE;
                }
              }
            }
            else { // Windows.
              if ( strpos($var, '/') || strpos($var, '\\') ) { // has slash, windows path cant have slash at pos zero
                if (($char = $var{0}) === '.') { // check for relative path
                  if (strpos($var, '../') === 0 || strpos($var, './') === 0
                        || strpos($var, '..\\') === 0 || strpos($var, '.\\') === 0) {
                    $pathRem = TRUE;
                  }
                }
                //  Absolute path (ascii letter + : + slash).
                //elseif (preg_match('/^[A-Za-z]\:[\/\x5C]/', $var)) {
                elseif ( ( ( ($char = ord($char)) >= 65 && $char <= 90) || ($char >= 97 && $char <= 122) )
                    && $var{1} === ':'
                    && (($char = $var{2}) === '/' || $char === '\\')
                ) {
                  $pathRem = TRUE;
                }
              }
            }
            if ($pathRem) {
              $var = basename($var);
            }
          }
          elseif ($le > self::$_docRootLength) {
            if (self::$_os_nix) { // non-windows
              if ($var{0} === '/' && ($le_noRoot = strlen($var = preg_replace(self::$_docRoot, '', $var))) < $le) {
                $pathRem = TRUE;
                $var = 'document_root' . $var;
                $truncedTo = $le = $le_noRoot;
              }
            }
            else {
              if (($le_noRoot = strlen($var = preg_replace(self::$_docRoot, '', $var))) < $le) {
                $pathRem = TRUE;
                $var = 'document_root' . $var;
                $truncedTo = $le = $le_noRoot;
              }
            }
          }
        }
        //  Escape string.
        $var = str_replace($options['needles'], $options['replacers'], $var);
        //  Truncate string.
        if ($le > $trunc) {
          $var = substr($var, 0, $trunc);
          $truncedTo = $trunc;
        }
        return $s
            . (!$truncedTo ? (!$pathRem ? '' : ('|-|!') ) :
                ('|' . $truncedTo . (!$pathRem ? '' : ('|!') ) ) )
            . ') '
            . $options['quote_begin'] . $var . $options['quote_end'];
      case 'resource':
        return '(resource)' . (!$options['hide_scalars'] ? (' ' . get_resource_type($var) ) : '');
      default: // NULL and unknown type
        return '(' . $t . ')';
    }
  }

  /**
   * Resolve constants.
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of the object's class
   * @param bool $keySort
   *   - default: falsy (~ do not sort constants, by name)
   * @return array|NULL
   *   - array, buckets: array to be inspected, message
   *   - FALSE on error
   */
  protected static function _constants($cat_or_instance_or_classname = 'user', $keySort = FALSE) {
    $cat = $cat_or_instance_or_classname;
    if (!$cat) {
      $a = get_defined_constants(TRUE);
      $ms = 'All defined constants';
    }
    elseif (is_string($cat)) {
      if (substr($cat, 0, 6) === 'class ') {
        if (($c = substr($cat, 6)) && class_exists($c)) {
          $o = new ReflectionClass($c);
          $a = $o->getConstants();
          $ms = 'Class ' . $c . ' constants';
        }
        else {
          return FALSE;
        }
      }
      else {
        $a = get_defined_constants($cat);
        $a = array_key_exists($cat, $a) ? $a[$cat] : array();
        $ms = 'Category ' . $cat . ' constants';
      }
    }
    elseif (is_object($cat)) {
      $c = get_class($cat);
      $o = new ReflectionClass($c);
      $a = $o->getConstants();
      $ms = 'Class ' . $c . ' constants';
    }
    else {
      return FALSE;
    }
    if ($keySort) {
      ksort($a);
    }
    return array($a, $ms);
  }

  /**
   * Log to database (currently watchdog).
   *
   * No permissions check for this method - everybody is allowed to log a message.
   * @param string $message
   *   - default: empty string
   * @param string $category
   *   - default: inspect
   * @param integer $severity
   *   - default: WATCHDOG_DEBUG
   * @return bool
   */
  public static function log_db($message = '', $category = 'inspect', $severity = WATCHDOG_DEBUG) {
    watchdog(
        $category,
        $message,
        NULL,
        $severity
    );
    return TRUE;
  }
  //  Utility methods.
  /**
   * Becomes the value of php:ini:precision.
   *
   * @var integer $_precision
   */
  protected static $_precision = 0;

  /**
   * Convert number to string avoiding E-notation for numbers outside system precision range.
   * @param float|integer $num
   * @return string
   */
  public static function numberToString($num) {
    if (!($prec = self::$_precision)) {
      $prec = self::$_precision = ini_get('precision');
    }
    return ($num < pow(10, $prec) && $num > pow(10, -$prec)) ?  // if within system precision, just string it
        ('' . $num) : number_format($num, 0, '.', '');
  }

  /**
   * Convert number or hex string to any base through 2 and 62  (~ larger than 16).
   *
   * Base 62 produces up to 35% shorter string than base 16.
   *
   *  Like native base_convert() precision fails for numbers (somewhere) larger than (tested on windows box with ini precision 14):
   *  - base 10: 9007199254740991 (> 15 digits)
   *  - base 16: 1fffffffffffff (> 14 digits)
   * @param integer|float|string $num
   * @param integer $fromBase
   *  - default: 10
   *  - only supports 10 and 16 (if other, returns arg $num unaltered)
   * @param integer $toBase
   *  - default (and max): 62
   * @return string
   */
  public static function baseConvert($num, $fromBase = 10, $toBase = 62) {
    $table = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ';
    $zero = $table{0};
    $remainder = $num;
    if($num < 1) {
      if(!$num) {
        return '' . $zero; // zero
      }
      $remainder *= -1;
    }
    if($toBase == $fromBase) {
      return $remainder;
    }
    switch ($fromBase) {
      case 10:
        break;
      case 16:
        $remainder = base_convert($remainder, 16, 10);
        break;
      default:
        return $remainder;
    }
    if($toBase > 62) {
      $toBase = 62;
    }
    else if($toBase < 2) {
      $toBase = 2;
    }
    $res = '';
    //  Find power ~ required number of digits.
    $pow = 0;
    $divisor = 1;
    $limit = 100; // probably way too high, havent seen more than 10 iterations (1fffffffffffff, with precision 14)
    while ($remainder >= ($test = pow($toBase, ++$pow)) && (--$limit)) {
      $divisor = $test;
      --$limit;
    }
    --$pow;
    //  Do modulo procedure.
    $limit = 100; // probably way too high, havent seen more than 14 iterations (1fffffffffffff, with precision 14)
    while ($remainder > 0 && (--$limit)) {
      if ($pow) {
        if ($divisor == $remainder) {
          return $res . $table{1} . str_repeat($zero, $pow);
        }
        else if ($divisor > $remainder) {
          $res .= $zero;
        }
        else {
          if ($remainder > $divisor) {
            $res .= $table{ floor($remainder / $divisor) };
            $remainder = fmod($remainder, $divisor);
          }
          else {
            $res .= $table{ $pow };
            $remainder = fmod($remainder, $divisor);
          }
          if($remainder < 1 && !floor($remainder)) {
            return $res . str_repeat($zero, $pow);
          }
        }
      }
      else {
        return $res . $table{ $remainder };
      }
      $divisor = pow($toBase, --$pow);
    }
    return $res;
  }
}