<?php
/**
 * @file
 * Drupal Inspect module stack trace inspection class
 */

/**
 * Stack trace inspector
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect */
final class InspectTrc extends InspectBs {

  /**
   * Default sub var recursion depth.
   *
   * Overrides InspectBs::DEPTH_DEFAULT - when methods of this class uses methods of InspectBs (using static::DEPTH_DEFAULT).
   *
   * @type integer
   */
  const DEPTH_DEFAULT = 0;

  /**
   * Minimum string truncation.
   *
   * Overrides InspectBs::TRUNCATE_MIN - when methods of this class uses methods of InspectBs (using static::TRUNCATE_MIN).
   *
   * @type integer
   */
  const TRUNCATE_MIN = 1000;

  /**
   * Default maximum stack frame depth
   *
   * @type integer
   */
  const TRACE_LIMIT_DEFAULT = 10;

  /**
   * Absolute maximum stack frame depth
   *
   * @type integer
   */
  const TRACE_LIMIT_MAX = 100;

  /**
   * Default trace options.
   *
   * Adds options that dont exist in InspectBs::$_optionsDefault
   *
   *   Formatting options:
   *   - (string) trace_spacer: spacer between frames
   *
   * @var array $_trace_optionsDefault
   */
  protected static $_trace_optionsDefault = array(
    'trace_spacer' => '- - - - - - - - - - - - - - - - - - - - - - - - -',
  );

  /**
   * Whether getting is desired and allowed
   * @var bool
   */
  protected $_get = FALSE;

  /**
   * Whether logging is desired and allowed
   * @var bool
   */
  protected $_log = FALSE;


  protected $_exceptionClass; // string
  protected $_trace; // array
  protected $_limit = 0;
  protected $_options_get; // array
  protected $_options_log; // array

  protected $_traceLog = '';
  protected $_traceGet = '';

  protected $_enclose_tag_get; // string
  protected $_enclose_tag_log; // string



  /**
   * Checks it's arguments against permissions.
   *
   * Admin user 1 is never allowed to _get_ traces (see InspectBs::_checkPermissionsAndOptions()).
   * @param Exception|falsy $xc
   *   - if falsy: create new trace
   * @param integer|array|falsy $limit_or_options
   * @param bool $get
   * @param bool $log
   * @return void
   */
  protected function  __construct($xc, $limit_or_options, $get = FALSE, $log = FALSE) {
    //  Get limit value if not array, and secure array having trace options.
    $opts = $limit_or_options;
    $limitOpt = $limitDef = static::TRACE_LIMIT_DEFAULT;
    $limitMax = static::TRACE_LIMIT_MAX;
    $optsDef =& static::$_trace_optionsDefault;
    if (!is_array($opts)) {
      if (!is_null($opts)) {
        $limitOpt = (int)$opts;
      }
      $opts = $optsDef; // copy, populate by $_trace_optionsDefault
    }
    else {
      if (isset($opts['limit'])) {
        $limitOpt = (int)$opts['limit'];
        unset($opts['limit']); // we dont need it any more
      }
      foreach ($optsDef as $k => $v) { // secure $_trace_optionsDefault buckets
        if (!isset($opts[ $k ])) {
          $opts[ $k ] = $v;
        }
      }
    }
    if ($limitOpt < 1 || $limitOpt > $limitMax) {
      $limitOpt = $limitMax;
    }
    $this->_limit = $limitOpt;
    //  Check permissions, and secure population of options array(s).
    if ($get && ($optsGet = static::_checkPermissionsAndOptions('get', $opts) ) ) {
      $this->_get = TRUE;
      $this->_enclose_tag_get = $optsGet['enclose_tag_get']; // store enclose tag
      $optsGet['enclose_tag_get'] = ''; // clear to avoid enclosure of every single var inspection
      $this->_options_get =& $optsGet;
    }
    if ($log && ($optsLog = static::_checkPermissionsAndOptions('log', $opts) ) ) {
      $this->_log = TRUE;
      $this->_enclose_tag_log = $optsLog['enclose_tag_log']; // store enclose tag
      $optsLog['enclose_tag_log'] = ''; // clear to avoid enclosure of every single var inspection
      $this->_options_log =& $optsLog;
    }

    if ($this->_get || $this->_log) {
      //  Create trace, if none given by arg.
      if ($xc) {
        if (is_object($xc) && (($u = get_class($xc)) === 'Exception' || is_subclass_of($xc, 'Exception'))) {
          $this->_exceptionClass = $u;
          $trace = $xc->getTrace();
          if (count($trace) > $limitOpt) {
            array_splice($trace, $limitOpt);
          }
          $this->_trace =& $trace;
        }
      }
      else {
        try {
          $trace = debug_backtrace();
        }
        catch (Exception $xc) {
          return;
        }
        if ($trace) {
          //  Remove top levels on synthetic trace, first of all the InspectTrc::__construct frame.
          array_shift($trace);
          //  If called via function instead of class method: remove class method call.
          if (isset($trace[0]['class']) && strtolower($trace[0]['class']) === 'inspecttrc'
              && isset($trace[1]['function']) && substr($trace[1]['function'], 0, 13) === 'inspect_trace') {
            array_shift($trace);
          }
          //  Enforce the trace limit.
          if (count($trace) > $limitOpt) {
            array_splice($trace, $limitOpt + 1); // plus one because we need the bucket holding the initial event
          }
          $this->_trace =& $trace;
        }
      }
      //  Format trace to get and/or log string.
      if ($this->_trace) {
        $this->_formatAsString($xc);
      }
    }
  }


  /**
   * Inspect and log stack trace to database (watchdog), if permitted.
   *
   *   Non-array value for $limit_or_options (limit):
   *   - maximum stack frame depth
   *   - non-NULL value is cast to integer
   *   - default: 10 (InspectTrc::TRACE_LIMIT_MAX)
   *   - max: 100 (InspectTrc::TRACE_LIMIT_MAX)
   *
   *   Array value for for $limit_or_options are like inspect()'s options, plus:
   *   - (integer) limit (default 10, max 100 ~ InspectTrc::TRACE_LIMIT_DEFAULT|TRACE_LIMIT_MAX)
   *   - (string) trace_spacer: spacer between frames (default hyphen dotted line, length 49)
   *
   * Default variable analysis depth (0 ~ InspectTrc::DEPTH_DEFAULT) is less than for plain variable inspection.
   *
   * String truncation minimum (1000 ~ InspectTrc::TRUNCATE_MIN) is less than for plain variable inspection.
   *
   * See inspect() for specifics of variable anylysis.
   *
   * watchdog severity: WATCHDOG_DEBUG
   *
   * Executes variable analysis within try-catch.
   * @see InspectVar::log()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param integer|array|falsy $limit_or_options
   *   - default: NULL
   * @return bool|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error
   */
  public static function log($exception = NULL, $limit_or_options = NULL) {
    $o = new InspectTrc($exception, $limit_or_options, FALSE, TRUE);
    //  If not allowed to log at all: get out.
    if (!$o->_log) {
      return NULL;
    }
    //  If tracing failed: get out.
    if (!$o->_traceLog) {
      return FALSE;
    }
    $tagStart = $tagEnd = '';
    if (($u = $o->_enclose_tag_log)) {
      $tagStart = '<' . $u . '>';  $tagEnd = '</' . $u . '>';
    }
    return self::log_db(
        $tagStart
          . self::_logNumber() . ' '
          . ( empty($o->_options_log['message']) ? '' : ($o->_options_log['message'] . ':' . $o->_options_log['delimiter']) )
          . '[Inspect trace - limit:' . $o->_limit . '|depth:' . $o->_options_log['depth'] . ']'
              . $o->_options_log['delimiter']
          . $o->_traceLog
          . $tagEnd,
        $o->_options_log['category']
    );
  }

  /**
   * Inspect stack trace and get the output as string, if permitted.
   *
   *   Behaviour and arguments like InspectTrc::log(), except for $limit_or_options options:
   *   - no enclose_tag_log options, use instead:
   *   - (string) enclose_tag_get: html enclosure tag for the log as a whole (default pre)
   * @see InspectTrc::log()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param integer|array|falsy $limit_or_options
   *   - default: NULL
   * @return string|bool
   *   - empty if user isnt permitted to get inspections
   *   - FALSE on error
   */
  public static function get($exception = NULL, $limit_or_options = NULL) {
    $o = new InspectTrc($exception, $limit_or_options, TRUE, FALSE);
    //  If not allowed to get at all: get out.
    if (!$o->_get) {
      return '';
    }
    //  If tracing failed: get out.
    if (!$o->_traceGet) {
      return FALSE;
    }
    $tagStart = $tagEnd = '';
    if (($u = $o->_enclose_tag_get)) {
      $tagStart = '<' . $u . '>';  $tagEnd = '</' . $u . '>';
    }
    return $tagStart
        . ( empty($o->_options_get['message']) ? '' : ($o->_options_get['message'] . ':' . $o->_options_get['delimiter']) )
        . '[Inspect trace - limit:' . $o->_limit . '|depth:' . $o->_options_get['depth'] . ']'
            . $o->_options_get['delimiter']
        . $o->_traceGet
        . $tagEnd;
  }

  /**
   * Convenience method - log and get stack trace inspection, if permitted.
   *
   * Behaviour and arguments like InspectTrc::log() and InspectTrc::get(),
   * and supports both $limit_or_options enclosure tag options.
   * @see InspectTrc::log()
   * @see InspectTrc::get()
   * @param Exception|falsy $exception
   *   - default: NULL (~ create new backtrace)
   *   - if Exception, trace that
   * @param integer|array|falsy $limit_or_options
   *   - default: NULL
   * @return string|bool
   *   - empty string if user isnt permitted to get inspections
   *   - FALSE on error
   */
  public static function log_get($exception = NULL, $limit_or_options = NULL) {
    $o = new InspectTrc($exception, $limit_or_options, TRUE, TRUE);
    //  If not allowed to get or log at all: get out.
    if (!$o->_get && !$o->_log) {
      return '';
    }
    //  If logging trace allowed, and tracing succeeded.
    if ($o->_log && $o->_traceLog) {
      $tagStart = $tagEnd = '';
      if (($u = $o->_enclose_tag_log)) {
        $tagStart = '<' . $u . '>';  $tagEnd = '</' . $u . '>';
      }
      self::log_db(
          $tagStart
            . self::_logNumber() . ' '
            . ( empty($o->_options_log['message']) ? '' : ($o->_options_log['message'] . ':' . $o->_options_log['delimiter']) )
            . '[Inspect trace - limit:' . $o->_limit . '|depth:' . $o->_options_log['depth'] . ']'
                . $o->_options_log['delimiter']
            . $o->_traceLog
            . $tagEnd,
          $o->_options_log['category']
      );
    }
    //  If getting trace allowed, and tracing succeeded.
    if ($o->_get && $o->_traceGet) {
      $tagStart = $tagEnd = '';
      if (($u = $o->_enclose_tag_get)) {
        $tagStart = '<' . $u . '>';  $tagEnd = '</' . $u . '>';
      }
      return $tagStart
          . ( empty($o->_options_get['message']) ? '' : ($o->_options_get['message'] . ':' . $o->_options_get['delimiter']) )
          . '[Inspect trace - limit:' . $o->_limit . '|depth:' . $o->_options_get['depth'] . ']'
              . $o->_options_get['delimiter']
          . $o->_traceGet
          . $tagEnd;
    }
    else if (!$o->_get) {
      return '';
    }
    return FALSE;
  }

  /**
   * Analyse and format the trace frames' buckets.
   * @param Exception|falsy $xc
   *   - default: NULL
   * @return void
   */
  protected function _formatAsString($xc = NULL) {
    try {
      //  Resolve options for get and/or log.
      if (($get = $this->_get)) {
        $sGet = '';
        $options_get =& $this->_options_get;
        $delim = $options_get['delimiter'];
        $frameSpacer = $options_get['trace_spacer'];
        $hidePaths_get = $this->_options_get['hide_paths'];
      }
      if (($log = $this->_log)) {
        $sLog = '';
        $options_log =& $this->_options_log;
        if (!$get) {
          $delim = $options_log['delimiter'];
          $frameSpacer = $options_log['trace_spacer'];
        }
        $hidePaths_log = $this->_options_log['hide_paths'];
      }
      //  If exception: resolve its origin.
      $file = $file_noRoot = $file_basename = 'unknown';
      if ($this->_exceptionClass) {
        $sLog = $sGet = 'Exception (' . $this->_exceptionClass . ') - code: ' . intval($xc->getCode());
        $file = trim($xc->getFile());
        $line = (($u = $xc->getLine()) ? trim($u) : '?');
        $message = 'message: ' . $xc->getMessage() . $delim;
      }
      else {
        $frame = array_shift($this->_trace);
        if (isset($frame['class'])) {
          $sLog = $sGet = ($frame['class'] . '::' . (isset($frame['function']) ? ($frame['function'] . '()') : '') );
        }
        else if(isset($frame['function'])) {
          $sLog = $sGet = ($frame['function'] . '()');
        }
        if (isset($frame['file'])) {
          $file = trim($frame['file']);
        }
        $line = (isset($frame['line']) ? trim($frame['line']) : '?');
        $message = '';
      }
      if (($file_length = strlen($file))) {
        if ( ($get && $hidePaths_get) || ($log && $hidePaths_log) ) {
          $file_basename = basename( $file );
        }
        if ( ($get && !$hidePaths_get) || ($log && !$hidePaths_log) ) {
          if (strlen($file_noRoot = preg_replace(self::$_docRoot, '', $file)) < $file_length) {
            $file_noRoot = 'document_root' . $file_noRoot;
          }
        }
      }
      if ($get) {
        $sGet .= $delim . 'file: ' . ($hidePaths_get ? $file_basename : $file_noRoot) . $delim
            . 'line: ' . $line . $delim
            . $message;
      }
      if ($log) {
        $sLog .= $delim . 'file: ' . ($hidePaths_log ? $file_basename : $file_noRoot) . $delim
            . 'line: ' . $line . $delim
            . $message;
      }
      //  Iterate stack frames.
      $nFrame = -1;
      foreach ($this->_trace as &$frame) {
        $s = (++$nFrame) . ' ' . $frameSpacer . $delim;
        if ($get) {
          $sGet .= $s;
        }
        if ($log) {
          $sLog .= $s;
        }
        //  File.
        $file = $file_noRoot = $file_basename = 'unknown';
        if (isset($frame['file']) && ($file_length = strlen($file = trim($frame['file'])))) {
          if ( ($get && $hidePaths_get) || ($log && $hidePaths_log) ) {
            $file_basename = basename( $file );
          }
          if ( ($get && !$hidePaths_get) || ($log && !$hidePaths_log) ) {
            if (strlen($file_noRoot = preg_replace(self::$_docRoot, '', $file)) < $file_length) {
              $file_noRoot = 'document_root' . $file_noRoot;
            }
          }
        }
        if ($get) {
          $sGet .= 'file: ' . ($hidePaths_get ? $file_basename : $file_noRoot);
        }
        if ($log) {
          $sLog .= 'file: ' . ($hidePaths_log ? $file_basename : $file_noRoot);
        }
        //  Line.
        $line = 'line: ' . (isset($frame['line']) ? trim($frame['line']) : '?') . $delim;
        if ($get) {
          $sGet .= ($hidePaths_get ? ' - ' : $delim) . $line;
        }
        if ($log) {
          $sLog .= ($hidePaths_log ? ' - ' : $delim) . $line;
        }
        //  Class, object, function, type.
        $sFunc = '';
        if (isset($frame['function'])) {
          if (strlen($u = $frame['function']) ) {
            $sFunc = $u;
          }
        }
        $sCls = '';
        if (isset($frame['class'])) {
          if (strlen($u = $frame['class'])) {
            $sCls = $u;
          }
        }
        $sType = '';
        if (isset($frame['type'])) {
          $sType = trim($frame['type']);
        }
        $sObj = '';
        if (isset($frame['object'])) {
          if (($o = $frame['object'])) {
            //  We dont know if class bucket is present when object is, but using class is cheaper.
            $sObj = $sCls ? $sCls : get_class($o);
            if ($sFunc) {
              $sFunc = 'method: (' . $sObj . ')' . ($sType ? $sType : '->') . $sFunc;
            }
            else { // Is this possible at all? Simply and object, no method call??
              $sObj = 'object (' . $sObj . ')';
            }
          }
        }
        elseif ($sFunc) {
          $sFunc = !$sCls ? ('function: ' . $sFunc) : ('static method: ' . $sCls . ($sType ? $sType : '::') . $sFunc);
          //  $frame['object'] doesnt exist.
        }
        /* else {
         *  $frame['object'] doesnt exist
         *  $frame['function'] may exist, containing non-empty string
         * }
         */
        if ($sFunc) {
          if ($get) {
            $sGet .= $sFunc . $delim;
          }
          if ($log) {
            $sLog .= $sFunc . $delim;
          }
        }
        elseif ($sObj) {
          if ($get) {
            $sGet .= $sObj . $delim;
          }
          if ($log) {
            $sLog .= $sObj . $delim;
          }
        }
        //  Args.
        if (isset($frame['args'])) {
          $le = count($frame['args']);
          $sArgsGet = '';
          if ($get) {
            $sArgsGet = '';
            //  If only analyzing args to depth zero, we want every arg on a line of its own.
            $comma = !$this->_options_get['depth'] ? (',' . $delim . ' ') : ', ';
            for ($i = 0; $i < $le; $i++) {
              $sArgsGet .= ($i ? $comma : '')
                  . self::_inspect(
                      $frame['args'][$i],
                      $options_get,
                      0 // recursion counter
                    );
            }
            $sGet .= 'args (' . $le . ')' 
                . (!$le ? '' : (':' . (!$this->_options_get['depth'] ? $delim : '') . ' ' . $sArgsGet) )
                . $delim;
          }
          if ($log) {
            $sArgsLog = '';
            //  If only analyzing args to depth zero, we want every arg on a line of its own.
            $comma = !$this->_options_log['depth'] ? (',' . $delim . ' ') : ', ';
            for ($i = 0; $i < $le; $i++) {
              $sArgsLog .= ($i ? $comma : '')
                  . self::_inspect(
                      $frame['args'][$i],
                      $options_log,
                      0 // recursion counter
                    );
            }
            $sLog .= 'args (' . $le . ')' 
                . (!$le ? '' : (':' . (!$this->_options_log['depth'] ? $delim : '') . ' ' . $sArgsLog) )
                . $delim;
          }
        }
      }
      if ($get) {
        $sGet .= 'END ' . $frameSpacer;
        $this->_traceGet =& $sGet;
      }
      if ($log) {
        $sLog .= 'END ' . $frameSpacer;
        $this->_traceLog =& $sLog;
      }
    }
    catch (Exception $xc0) {
    }
  }
}