<?php
/**
 * @file
 * Drupal Inspect module variable inspection class
 */

/**
 * Variable inspector
 * @static
 * @singleton
 * @category Drupal
 * @package Developer
 * @subpackage Inspect */
final class InspectVar extends InspectBs {
  /**
   * Prevents erratic instantiation of this entirely static class
   */
  private function  __construct() {}

  /**
   * Inspect variable and log to database (watchdog), if permitted.
   *
   *   Like PHP native var_dump()/var_export() except:
   *   - is controlled by permissions
   *   - has much more options, and far better formatting (JSONesque)
   *   - doesnt fail on inspecting $GLOBALS
   *   - doesnt echo, logs to database log
   *
   *   Non-array value for $depth_or_options (depth):
   *   - the value controls how deeply the variable is going to be analyzed
   *   - non-NULL value is cast to integer
   *   - default: 10 (InspectBs::DEPTH_DEFAULT)
   *   - max: 10 (InspectBs::DEPTH_MAX)
   *
   *   Array value for for $depth_or_options (options, any number of):
   *   - (integer) depth (default 10, max 10 ~ InspectBs::DEPTH_DEFAULT|DEPTH_MAX)
   *   - (bool) hide_scalars: hide values of scalar vars, and type of resource (default FALSE, truthy cannot override permissions)
   *   - (integer) truncate (default 100, max 10000 ~ InspectBs::TRUNCATE_MIN|TRUNCATE_DEFAULT)
   *   - (bool) hide_paths: only relevant for log (paths always hidden for get), and falsy cannot override permissions
   *   - (string) message (default empty)
   *   - (string) category: logging category (default inspect)
   *   - (string|array) filter: filter out that|those key name(s) (default empty)
   *   - (array) needles: list of char|string needles to be used in str_replace (default \n \r < > " ')
   *   - (array) replacers: list of char|string replacers to be used in str_replace (default equivalent xml entities)
   *   - (string) quote_begin: char|string used as initial quote (default >>)
   *   - (string) quote_end: char|string used as final quote (default <<)
   *   - (string) delimiter: between object/array buckets (default carriage return)
   *   - (string) indent: indentation of object/array buckets (default 2 spaces)
   *   - (string) enclose_tag_log: html enclosure tag for the log as a whole (default pre)
   *
   *   String truncation:
   *   - truncate option less than one (hide strings) doesnt apply to strings that solely consist of digits (stringed integers)
   *   - a string bucket of an object/array keyed pass or password (lowercase) will always be hidden (no matter it's content)
   *
   *   Removes paths at the beginning of variables:
   *   - always removes document root
   *   - always removes relative and absolute paths when getting (optional when logging)
   *   - if *nix server OS: only strings having no space(s), unix style absolute paths, and only looks for forward slash
   *   - if windows server OS: windows style absolute paths, and looks for forward and backward slashes
   *   - path removal takes place before escaping and truncating (except for overly long strings)
   *
   *   Reports non-integer numbers:
   *   - double as float
   *   - NaN as NaN and infinite as infinite
   *
   *   Reports string lengths and path removal (path reduced to filename):
   *   - un-truncated string, and when hiding scalar values: (multibyte length|ascii length)
   *   - truncated string: (multibyte length|ascii length|truncation length)
   *   - if path removed: there will be a fourth flag (exclamation mark), if no truncation truncation length will be hyphen
   *
   *   Recursive:
   *   - always stops at 10th (InspectBs::DEPTH_MAX) recursion or less (~ the depth option)
   *   - when buckets of an object/array is analyzed, buckets of buckets will be analyzed to one lesser depth
   *   - detects if a bucket of an object references the object itself (marked *RECURSION*)
   *   - detects GLOBALS array's self-referencing GLOBALS bucket (marked *RECURSION*)
   *   - doesnt check identity for arrays (not possible in PHP), nor 'same-ness' (too heavy, and prone to err)
   *
   * watchdog severity: WATCHDOG_DEBUG
   *
   * Executes variable analysis within try-catch.
   * @param mixed $var
   * @param integer|array|falsy $depth_or_options
   *   - default: NULL
   * @return bool|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error
   */
  public static function log($var, $depth_or_options = NULL) {
    if (!($opts = self::_checkPermissionsAndOptions('log', $depth_or_options)) ) {
      return NULL;
    }
    try {
      $nspct = self::_inspect($var, $opts);
    }
    catch (Exception $xc) {
      return FALSE;
    }
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag_log'])) {
      $tagStart = '<' . $u . '>';  $tagEnd = '</' . $u . '>';
    }
    return self::log_db(
        $tagStart
          . self::_logNumber() . ' '
          . ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $opts['delimiter']) )
          . '[Inspect - depth:' . $opts['depth'] . ']' . $opts['delimiter']
          . $nspct
          . $tagEnd,
        $opts['category']
    );
  }

  /**
   * Inspect variable and get the output as string, if permitted.
   *
   *   Behaviour and arguments like inspect(), except for $depth_or_options options:
   *   - no enclose_tag_log options, use instead:
   *   - (string) enclose_tag_get: html enclosure tag for the output as a whole (default pre)
   * @param mixed $var
   * @param integer|array|falsy $depth_or_options
   *   - default: NULL
   * @return string|bool
   *   - empty string if user isnt permitted to get inspections
   *   - FALSE on error
   */
  public static function get($var, $depth_or_options = NULL) {
    if (!($opts = self::_checkPermissionsAndOptions('get', $depth_or_options)) ) {
      return '';
    }
    try {
      $nspct = self::_inspect($var, $opts);
    }
    catch (Exception $xc) {
      return FALSE;
    }
    $tagStart = $tagEnd = '';
    if (($u = $opts['enclose_tag_get'])) {
      $tagStart = '<' . $u . '>';  $tagEnd = '</' . $u . '>';
    }
    return $tagStart
        . ( empty($opts['message']) ? '' : ($opts['message'] . ':' . $opts['delimiter']) )
        . '[Inspect - depth:' . $opts['depth'] . ']' . $opts['delimiter']
        . $nspct . $tagEnd;
  }

  /**
   * Convenience method - log and get variable inspection, if permitted.
   *
   * Behaviour and arguments like InspectVar::log() and InspectVar::get(), and supports both $depth_or_options enclosure tag options.
   * @see InspectVar::log()
   * @see InspectVar::get()
   * @param mixed $var
   * @param integer|array|falsy $depth_or_options
   *   - default: NULL
   * @return string|bool
   *   - empty string if user isnt permitted to get inspections
   *   - FALSE on error
   */
  public static function log_get($var, $depth_or_options = NULL) {
    //  Has to create two inspections, because permissions may differ.
    self::log($var, $depth_or_options);
    return self::get($var, $depth_or_options);
  }

  /**
   * Inspect constants and log to database (watchdog), if permitted.
   *
   *   $options options like InspectVar::log(), except:
   *   - (bool) sort: if truthy, sorts constants by name (except if inspecting all defined constants, then only category is sorted)
   * @see InspectVar::log()
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of the object's class
   * @param array|falsy $options
   *   - default: NULL
   * @return bool|NULL
   *   - NULL if user isnt permitted to log inspections
   *   - FALSE on error
   */
  public static function constants_log($cat_or_instance_or_classname = 'user', $options = NULL, $_get = FALSE) {
    $sort = FALSE;
    $ms = TRUE;
    if ($options && is_array($options)) {
      if (array_key_exists('sort', $options) && $options['sort']) {
        $sort = TRUE;
      }
      unset($options['sort']);
      if (array_key_exists('message', $options) && $options['message']) {
        $ms = FALSE;
      }
    }
    else {
      $options = array();
    }
    if (!($a = self::_constants($cat_or_instance_or_classname, $sort))) {
      return FALSE;
    }
    if ($ms) {
      $options['message'] = $a[1];
    }
    return !$_get ? self::log($a[0], $options) : self::get($a[0], $options);
  }

  /**
   * Inspect constants and get the output as string, if permitted.
   *
   *   $options options like InspectVar::get(), except:
   *   - (bool) sort: if truthy, sorts constants by name (except if inspecting all defined constants, then only category is sorted)
   * @see InspectVar::get()
   * @param string|object|falsy $cat_or_instance_or_classname
   *   - default: 'user' (~ the user category of defined constants)
   *   - falsy: all defined constants, listed by category
   *   - 'class SomeClassName': class constants of class SomeClassName
   *   - other strings: that category of defined constants
   *   - object: class constants of the object's class
   * @param array|falsy $options
   *   - default: NULL
   * @return string|bool
   *   - empty string if user isnt permitted to get inspections
   *   - FALSE on error
   */
  public static function constants_get($cat_or_instance_or_classname = 'user', $options = NULL) {
    return self::constants_log($cat_or_instance_or_classname, $options, TRUE);
  }
}